@@ -1,8 +1,6 @@
-from .base import get_model, create_model
+from nowcasting_utils.models.base import get_model, create_model
 from .conv_lstm import EncoderDecoderConvLSTM, ConvLSTM
 from .pl_metnet import LitMetNet
-
 from .runet import R2U_Net, RUnet
 from .attention_unet import R2AttU_Net, AttU_Net
-
 from .perceiver import Perceiver
